/**
 * A personnel register maintains a collection of employees.  A company
 * can use a personnel register to keep track of who is employed in the
 * company.
 *
 * The personnel register collects gender statistics for all the
 * employees that are added to the register.  These statistics reflect
 * the gender of each employee at the time the employee is added.  The
 * statistics will not reflect any gender change done later.
 *
 * The personnel register has some major limitations: the register will
 * stop working if the number of employees exceeds 100, and there is no
 * way to remove employees from the register, once they are added.
 */
public class PersonnelRegister {
  static final int MAXIMUM_NUMBER_OF_EMPLOYEES = 100;
  private Employee[] allEmployees;
  private int employeeCount;
  private int womenCount;

  public PersonnelRegister() {
    allEmployees = new Employee[MAXIMUM_NUMBER_OF_EMPLOYEES];
  }

  /**
   * Stores another employee in the register and updates the gender
   * statistics.  Only an employee with known gender can be added
   *
   * @return The ID number assigned to the employee that was added.
   */
  public int addEmployee(Employee newEmployee) {
    assert newEmployee.gender != Gender.UNKNOWN
      : "Gender must be known before registration.";
    assert employeeCount < allEmployees.length;

    int employeeIdNumber = employeeCount;
    allEmployees[employeeIdNumber] = newEmployee;
    addEmployeeToStatistics(newEmployee);
    return employeeIdNumber;
  }

  private void addEmployeeToStatistics(Employee newEmployee) {
    employeeCount++;
    if (newEmployee.gender == Gender.FEMALE) {
      womenCount++;
    }
  }

  /**
   * Retrieves a stored employee with a specific ID number.
   *
   * @param employeeIdNumber The ID number of the employee to retrieve.  A
   *     valid ID number must be non-negative and less than the number
   *     returned by numberOfEmployees().
   */
  public Employee getEmployee(int employeeIdNumber) {
    return allEmployees[employeeIdNumber];
  }

  /**
   * Retrieves a stored employee with a specific name.
   *
   * @param firstName The first name of the employee.
   * @param lastName The last name of the employee.
   * @return the employee, or null if no employee with that name was
   *     found.
   */
  public Employee getEmployee(String firstName, String lastName) {
    for (int i = 0; i < employeeCount; i++) {
      Employee anEmployee = allEmployees[i];
      if (firstName.equals(anEmployee.firstName) &&
          lastName.equals(anEmployee.lastName)) {
        return anEmployee;
      }
    }
    return null;
  }

  /**
   * @return The total number of employees that have been added to
   *         this register.
   */
  public int numberOfEmployees() {
    return employeeCount;
  }

  /**
   * @return The percentage of women employees in the register.  Since
   *     the register keeps track of men and women added to the
   *     register, it is possible to calculate the percentage of men
   *     using this formula: 100 - women_percentage.
   */
  public double percentageWomen() {
    return 100.0 * womenCount / employeeCount;
  }
}