/**
 * An employee is paid by the hour at a fixed hourly rate.
 * However, for each hour worked overtime during the week,
 * the hourly rate is doubled.
 */
class Employee {
  // Static variable (constant)
  public final static double NORMAL_WORKWEEK = 37.5;

  // Field variables
  String firstName;
  String lastName;
  double hourlyRate;

  // Constructors
  /**
   * Creates an employee with the given name and the hourly rate.
   * @param firstName First name (and middle names, if any) of the employee.
   * @param lastName Last name of the employee.
   * @param hourlyRate Payment per hour in GBP.
   */
  public Employee(String firstName, String lastName, double hourlyRate) {
    this.firstName = firstName;
    this.lastName = lastName;
    this.hourlyRate = hourlyRate;
  }

  // Instance methods
  /**
   * Calculates the weekly salary for the employee.
   * @param numHours Number of hours worked in current week.
   * @return Weekly salary in GBP for the current week.
   *
   * @pre numHours >= 0.0
   * @post salary >= 0.0
   */
  double calculateSalary(double numHours) {
    assert numHours >= 0.0 : "Number of hours must be >= 0.0";       // @pre
    double salary = hourlyRate * NORMAL_WORKWEEK;
    if (numHours > NORMAL_WORKWEEK) {
      salary += 2.0 * hourlyRate * calculateOvertime(numHours);
    }
    assert salary >= 0.0 : "Weekly salary must be >= 0.0";           // @post
    return salary;
  }

  // Overridden methods from the Object class.
  /**
   * Compares two objects for equality, i.e the current object and
   * the object whose reference value is passed in the parameter.
   * This method overrides the method in the Object class.
   * @param other The object the current object will be compared with.
   * @return true If the two objects have the same state; otherwise false.
   */
  public boolean equals(Object other) {
    if (other == this) return true;                // Same object?
    if (!(other instanceof Employee)) return false;// dynamic type control
    Employee emp = (Employee) other;               // downcasting
    return firstName.equals(emp.firstName)         // using equals() in String
           && lastName.equals(emp.lastName)
           && hourlyRate == emp.hourlyRate;
  }

  /**
   * Creates a string representation of an employee.
   * This method overrides the method in the Object class.
   * @return A string representing the current state of the object.
   */
  public String toString() {
    return String.format(
                  "First name: %s\tLast name: %s\tHourly rate: %.2f GBP",
                  firstName, lastName, hourlyRate);
  }

  // Static methods
  /**
   * Calculates number of hours overtime an employee has worked this week.
   * @param numHours Total number of hours worked by an employee.
   * @return Number of hours worked overtime.
   *
   * @pre numHours >= 0.0
   * @post return >= 0.0
   */
  public static double calculateOvertime(double numHours) {
    assert numHours >= 0.0 : "Number of hours must be >= 0.0";        // @pre
    double overtime = 0.0; // assume no overtime
    if (numHours > NORMAL_WORKWEEK) {
      overtime = numHours - NORMAL_WORKWEEK;
    }
    assert overtime >= 0.0:"Number of overtime hours must be >= 0.0"; // @post
    return overtime;
  }
}