/**
 * A manager is paid by the hour at a fixed hourly rate.
 * For each hour worked overtime during the week,
 * the hourly rate is doubled.
 * In addition, the manager always receives a fixed manager bonus.
 * The manager is also a union member.
 */
class ManagerV3 extends EmployeeV0 implements IUnionMember {
  // Field variable
  double managerBonus;

  // Constructors
  /**
   * Creates a manager with default values for all fields.
   */
  ManagerV3() { // explicit default constructor
    super();                                                         // (1)
  }
  /**
   * Creates a manager with the given name and hourly rate.
   * The manager bonus is set to the default value (0.0).
   * @param firstName First name (and middle names, if any) of the manager.
   * @param lastName Last name of the manager.
   * @param hourlyRate Payment per hour in GBP.
   */
  ManagerV3(String firstName, String lastName, double hourlyRate) {
    super(firstName, lastName, hourlyRate);                          // (2)
  }
  /**
   * Creates a manager with the given name, hourly rate and manager bonus.
   * @param firstName First name (and middle names, if any) of the manager.
   * @param lastName Last name of the manager.
   * @param hourlyRate Payment per hour in GBP.
   * @param managerBonus A bonus added to the weekly salary.
   */
  ManagerV3(String firstName, String lastName,
            double hourlyRate, double bonus) {
    super(firstName, lastName, hourlyRate);                          // (3)
    managerBonus = bonus;                          // Stores extended state
  }

  // Instance methods
  /**
   * Calculates the weekly salary of a manager.
   * This method overrides the method in the EmployeeV0 class.
   * @param numHours Number of hours worked in current week.
   * @return Weekly salary in GBP for the current week.
   */
  double calculateSalary(double numHours) {                          // (4)
    // Using assert to check the value of the actual parameter
    assert numHours >= 0 : "Number of hours must be >= 0";
    double fixedSalary = super.calculateSalary(numHours);            // (5)
    fixedSalary += managerBonus;                                     // (6)
    return fixedSalary;
  }
  /**
   * Prints the name, hourly rate and bonus for the manager.
   * This method overrides the method in the EmployeeV0 class.
   */
  void printState() {                                                // (7)
    super.printState();
    System.out.printf("Manager bonus: %.2f GBP%n", managerBonus);
  }

  // Implements the IUnionMember interface
  /**
   * Calculates the union fee based on weekly salary and employee category.
   */
  public double calculateFee(double weeklySalary, int empCategory) { // (8)
    assert weeklySalary >= 0.0 : "Weekly salary must be >= 0.0";     // @pre
    assert
      empCategory == SUBORDINATE ||
      empCategory == MIDDLE_MANAGEMENT ||
      empCategory == UPPER_MANAGEMENT: "Invalid employee category";  // @pre
    double fee = 0.0;
    switch (empCategory) {
      case SUBORDINATE:
        fee = 0.05 * weeklySalary;
        break;
      case MIDDLE_MANAGEMENT:
        fee = 0.10 * weeklySalary;
        break;
      case UPPER_MANAGEMENT:
        fee = 0.15 * weeklySalary;
        break;
    }
    assert fee >= 0.0 : "Membership fee must be >= 0.0";             // @post
    return fee;
  }
}