// File: GameBoard.java
enum Direction {
  UP(0, 1), LEFT(1, 0), DIAGONAL1(1, 1), DIAGONAL2(1, -1),
  DOWN(-UP.columnStep,-UP.rowStep),
  RIGHT(-LEFT.columnStep,-LEFT.rowStep),
  OPPOSITE_DIAGONAL1(-DIAGONAL1.columnStep,-DIAGONAL1.rowStep),
  OPPOSITE_DIAGONAL2(-DIAGONAL2.columnStep,-DIAGONAL2.rowStep);

  final static Direction[] DIRECTIONS
               = {UP, LEFT, DIAGONAL1, DIAGONAL2};
  final static Direction[] OPPOSITE_DIRECTIONS
               = {DOWN, RIGHT, OPPOSITE_DIAGONAL1, OPPOSITE_DIAGONAL2};

  int columnStep;
  int rowStep;

  Direction(int dx, int dy) {
    columnStep = dx;
    rowStep = dy;
  }

  Direction opposite() {
    return OPPOSITE_DIRECTIONS[this.ordinal()];
  }
}

public class GameBoard {
  public static final int NO_VACANT_ROW = -1;
  public static final char EMPTY_CELL = '\u0000';
  public static final int ROW_COUNT = 6;
  public static final int COLUMN_COUNT = 7;
  public static final int CELL_COUNT = ROW_COUNT * COLUMN_COUNT;

  char[][] cellContents = new char[COLUMN_COUNT][ROW_COUNT];
  int moveCounter;
  char winner;
  IPlayer[] players;

  GameBoard() {
    assert cellContents[0][0] == EMPTY_CELL;
  }

  void setPlayers(IPlayer[] bothPlayers) {
    assert bothPlayers.length == 2;
    players = bothPlayers;
  }

  char pieceAt(int column, int row) {
    return cellContents[column][row];
  }

  int findBottommostVacantRow(int column) {
    char[] columnContents = cellContents[column];
    for (int row = 0; row < columnContents.length; ++row) {
      if (columnContents[row] == EMPTY_CELL)
        return row;
    }
    return NO_VACANT_ROW;
  }

  boolean hasWinner() {
    return winner != '\u0000';
  }

  boolean isGameOver() {
    return (moveCounter >= CELL_COUNT) || hasWinner();
  }

  void dropPieceDownColumn(char piece, int column) {
    assert !isGameOver();
    int row = findBottommostVacantRow(column);
    cellContents[column][row] = piece;
    if (longestSequenceContainingCell(column, row) >= 4)
      winner = piece;
    ++moveCounter;
  }

  int longestSequenceContainingCell(int column, int row) {
    int largestLength = 1;
    // Look for sequences of similar pieces in all directions.
    for (Direction direction : Direction.DIRECTIONS) {
      int piecesInALine = 1 + numOfConsecutivePieces(column, row, direction)
          + numOfConsecutivePieces(column, row, direction.opposite());
      largestLength = Math.max(largestLength, piecesInALine);
    }
    return largestLength;
  }

  int numOfConsecutivePieces(int column, int row, Direction direction) {
    char piece = pieceAt(column, row);
    int count = 0;
    while (true) {
      column += direction.columnStep;
      row += direction.rowStep;
      if (!positionContainsPiece(column, row, piece))
        break;
      ++count;
    }
    return count;
  }

  boolean validColumnIndex(int column) {
    return (column >= 0) && (column < COLUMN_COUNT);
  }

  boolean validRowIndex(int row) {
    return (row >= 0) && (row < ROW_COUNT);
  }

  boolean positionContainsPiece(int column, int row, char piece) {
    return validColumnIndex(column)
        && validRowIndex(row)
        && (pieceAt(column, row) == piece);
  }

  boolean isColumnFull(int column) {
    char[] columnContents = cellContents[column];
    int indexOfTopmostCell = columnContents.length - 1;
    char topmostCell = columnContents[indexOfTopmostCell];
    return topmostCell != EMPTY_CELL;
  }

  boolean isValidColumnSelection(int column) {
    return validColumnIndex(column)
        && !isColumnFull(column);
  }

  char nextPiece() {
    boolean isEvenNumberedMove = moveCounter % 2 == 0;
    if (isEvenNumberedMove) {
      return 'X';
    } else {
      return 'O';
    }
  }

  IPlayer nextPlayer() {
    return players[moveCounter % 2];
  }
}