import java.util.Iterator;
/**
 * This class implements a linked list, allowing nodes to be inserted and
 * removed.
 */
public class LinkedList<E> implements ILinkedList<E> {

  /** Head of the list */
  private Node<E> head;

  /** Tail of the list */
  private Node<E> tail;

  /** Number of nodes in the list */
  private int numberOfNodes;

  // Mutators
  /**
   * Insert a node at the head of the list.
   * @param dataObj Data to be stored in the list.
   */
  public void insertAtHead(E dataObj) {                        // (1)
    Node<E> p = new Node<E>(dataObj, null);
    if (isEmpty()) {
      head = tail = p;
    } else { // head = new MyNode<E>(dataObj, head);
      p.setNext(head);   // Step 1
      head = p;          // Step 2
    }
    numberOfNodes++;
  }

  /**
   * Insert a node at the tail of the list.
   * @param dataObj Data to be stored in the node.
   */
  public void insertAtTail(E dataObj) {                         // (2)
    Node<E> p = new Node<E>(dataObj, null);
    if (isEmpty()) {
      head = tail = p;
    } else {
      tail.setNext(p); // Step 1
      tail = p;        // Step 2
    }
    numberOfNodes++;
  }

  /**
   * Remove first node in the list.
   * Assume the list is not empty.
   * @return Data in the node which is removed.
   */
  public E removeFromHead() {                                   // (3)
    assert numberOfNodes > 0 : "The list cannot be empty.";
    Node<E> p = head;         // Step 1
    if (head == tail) {
      head = tail = null;     // Only one node in the list.
    } else {
      head = p.getNext();     // Step 2
    }
    numberOfNodes--;
    return p.getData();
  }

  /**
   * Remove last node in the list.
   * Assume the list is not empty.
   * @return Data in the node which is removed.
   */
  public E removeFromTail() {                                   // (4)
    assert numberOfNodes > 0 : "The list cannot be empty.";
    Node<E> p = tail;              // Step 1
    if (head == tail) {
      head = tail = null;          // Only one node in the list.
    } else {
      Node<E> nextToLast = head;   // Step 2
      while (nextToLast.getNext() != tail) {
        nextToLast = nextToLast.getNext();
      }
      tail = nextToLast;           // Step 3
      tail.setNext(null);          // Step 4
    }
    numberOfNodes--;
    return p.getData();
  }

  /**
   * Remove node with given data from the list.
   * @param dataObj Data in the node which is removed.
   * @return True, if a node with the given data is removed.
   */
  public boolean remove(E dataObj) {                            // (5)
    // Traverse the list.
    Node<E> currentNode = head;
    Node<E> predecessor = null;
    boolean found = false;
    while (currentNode != null && !found) {
      if (currentNode.getData().equals(dataObj)) {
        found = true;
      } else {
        // Not yet found. Update predecessor and currentNode references.
        predecessor = currentNode;
        currentNode = currentNode.getNext();
      }
    }
    if (found) {
      if (predecessor == null) {
        // Found in first node of the list.
        removeFromHead();
      } else if (currentNode.getNext() == null) {
        // Found in last node of the list.
        tail = predecessor;
        tail.setNext(null);
        numberOfNodes--;
      } else {
        // Found inside the list.
        predecessor.setNext(currentNode.getNext());
        numberOfNodes--;
      }
    }
    return found;
  }

  // Selectors
  /**
   * Determines whether the list is empty.
   * @return True, if the list is empty, otherwise return false.
   */
  public boolean isEmpty() { return head == null; }

  /**
   * Get the number of nodes in the list.
   * @return Number of nodes in the list.
   */
  public int numberOfNodes() { return numberOfNodes; }

  /**
   * Get the first node in the list.
   * @return The first node in the list, if any.
   */
  public Node<E> first() { return head; }

  /**
   * Get the last node in the list.
   * @return The last node in the list.
   */
  public Node<E> last() { return tail; }

  /**
   * Find the node with the given data.
   * @param dataObj Data in the node to search for.
   * @return MyNode with the given data, otherwise return null.
   */
  public Node<E> find(E dataObj) {
    // Traverse the list, starting from the head.
    Node<E> currentNode = head;
    while (currentNode != null) {
      if (currentNode.getData().equals(dataObj)) {
        return currentNode;       // Found.
      }
      currentNode = currentNode.getNext();
    }
    return null;               // Not found.
  }

  // Other help methods
  /**
   * Make an iterator.
   * @return the iterator
   */
  public Iterator<E> iterator() {                               // (6)
    return new LinkedListIterator<E>(this);
  }

  /** Fill the array from the list. */
  public void listToArray(E[] toArray) {                        // (7)
    assert toArray.length == numberOfNodes : "Wrong size of array.";
    int i = 0;
    for (E data : this) {
      toArray[i++] = data;
    }
  }
}