import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;

class CompanyUsingRandomAccess {

  final static char FIELD_TERMINATOR    = ',';

  //  Constants for handling records with fixed length.                  (6)
  final static char PAD_CHAR             = '\u0000';
  final static int  FIRSTNAME_LENGTH     = 20; // Unicode character
  final static int  LASTNAME_LENGTH      = 30; // Unicode character
  final static int  HOURLYRATE_LENGTH    = 8;  // bytes
  final static int  GENDER_LENGTH        = 1;  // byte
  final static int  RECORD_LENGTH_BYTES  = 2*FIRSTNAME_LENGTH +
                                           2*LASTNAME_LENGTH +
                                             HOURLYRATE_LENGTH +
                                             GENDER_LENGTH; // bytes
  // Random access file.
  private RandomAccessFile raf;

  // Register with the employees.
  private PersonnelRegister register;

  // Constructors
  CompanyUsingRandomAccess() {
    register = new PersonnelRegister();
  }
  CompanyUsingRandomAccess(PersonnelRegister register) {
    this.register = register;
  }

  // Print statistics.
  void printReport() {
    System.out.println(register);
  }
  void printAllEmployeesToTerminalWindow() {
    Employee[] employees = register.getEmployeeArray();
    int numOfEmployees = register.getNumOfEmployees();
    System.out.println("No. of employees: " + numOfEmployees);
    for (int i = 0; i < numOfEmployees; i++) {
      System.out.println(employees[i]);
    }
  }

  /** Method opens a file for reading and writing. */
  void initRandomAccessFile(String dataFileName) {                    // (7)

    try {
      // Open file for reading and writing.
      raf = new RandomAccessFile(dataFileName, "rw");
      // Reset the file pointer.
      raf.setLength(0);
      // Reset counter for number of records on the file.
      raf.writeInt(0);
    }
    catch (FileNotFoundException exception) {
      System.out.print("File not found: " + exception);
      System.exit(1);
    }
    catch (IOException exception) {
      System.out.print("Error writing to file: " + exception);
      System.exit(1);
    }
  }

  /** Method appends an Employee record to the file. */
  void appendFileWithEmployee(Employee employee) {                    // (8)
    try {
      // Position file pointer after the last record.
      raf.seek(raf.length());
      // Write the employee data to the file.
      writeEmployeeData(employee);

      // Update the number of records in the file.
      raf.seek(0);
      int numOfEmployeesRegistered = raf.readInt();
      raf.seek(0);
      raf.writeInt(numOfEmployeesRegistered + 1);
    }
    catch (IOException exception) {
      System.out.print("Error writing to file: " + exception);
    }
  }

  /** Method writes field values of an employee. */
  void writeEmployeeData(Employee employee) throws IOException {
    writeFixedLengthString(employee.firstName, FIRSTNAME_LENGTH);
    writeFixedLengthString(employee.lastName, LASTNAME_LENGTH);
    raf.writeDouble(employee.hourlyRate);
    raf.writeByte(employee.gender.ordinal());
  }

  /** Method writes a string (padding, if necessary) of max. length. */
  private void writeFixedLengthString(String str, int maxFieldLength) // (9)
                                     throws IOException {
    //  Field padded with PAD_CHAR if the length
    //  is less than max length.
    int numOfPadChars = maxFieldLength - str.length();
    StringBuilder padding = new StringBuilder();
    for (int k = 0; k < numOfPadChars; k++) {
      padding.append(PAD_CHAR);
    }
    raf.writeChars(str + padding);
  }

  /** Method changes the hourly rate of an Employee record in the file. */
  void registerNewHourlyRateForEmployeeOnFile(int employeeNumber,
                                              double hourlyRate) {    // (10)
    try {
      // Set file pointer to the correct position.
      raf.seek(4 + employeeNumber * RECORD_LENGTH_BYTES);
      // Read the employee record and update hourly rate.
      Employee employee = readEmployeeData();                         // (11)
      employee.hourlyRate = hourlyRate;

      // Position file pointer to rewrite the updated employee record.
      long currentPosition = raf.getFilePointer();
      raf.seek(currentPosition - RECORD_LENGTH_BYTES);
      writeEmployeeData(employee);
    }
    catch (IOException exception) {
      System.out.print("Error reading from file: " + exception);
    }
  }

  /** Method reads an employee record from current position in the file. */
  Employee readEmployeeData() throws IOException {
    String firstName = readFixedLengthString(FIRSTNAME_LENGTH);
    String lastName = readFixedLengthString(LASTNAME_LENGTH);
    double hourlyRate = raf.readDouble();
    Gender gender = null;
    switch (raf.readByte()) { // Read ordinal value for gender.
      case 0: default:
        gender = Gender.FEMALE;
        break;
      case 1:
        gender = Gender.MALE;
        break;
    }
    return new Employee(firstName, lastName, hourlyRate, gender);
  }

  /** Method reads a string of specified max. length. */
  private String readFixedLengthString(int maxFieldLength)
                 throws IOException {                                 // (12)
    StringBuilder strBuffer = new StringBuilder();
    int i;
    for (i = 0; i < maxFieldLength; i++) {
      char character = raf.readChar();
      if (character == PAD_CHAR)
        break;
      else
        strBuffer.append(character);
    }
    //  Skip any PAD_CHARs, each of size 2 bytes.
    raf.skipBytes(2*(maxFieldLength - i - 1));
    return strBuffer.toString();
  }

  /** Method reads all employee records. */
  void readAllEmployeesFromRandomAccessFile() {                       // (13)
    try {
      // Set file pointer to beginning of file.
      raf.seek(0);

      // Create a new personnel register.
      register = new PersonnelRegister();

      // Read how many employees are in the file.
      int numOfEmployees = raf.readInt();

      // Read each employee's data sequentially.
      for (int i = 0; i < numOfEmployees; i++) {
        // Read an employee record.
        Employee employee = readEmployeeData();
        // Insert employee in the register.
        register.registerEmployee(employee);
      }
    }
    catch (IOException exception) {
      System.out.print("Error reading from file: " + exception);
      System.exit(1);
    }
  }

  /** Method closes files for random access. */
  void closeStream() {
    try {
      if (raf != null)
        raf.close();
    }
    catch (IOException exception) {
      System.out.print("Error on closing: " + exception);
      System.exit(1);
    }
  }

  /** Method changes hourly rate of an employee and also updates the file. */
  void setHourlyRate(int employeeNumber, double hourlyRate) {
    try {
      register.getEmployee(employeeNumber).hourlyRate = hourlyRate;
      writeEmployeeToRandomAccessFile(employeeNumber);
    }
    catch (ArrayIndexOutOfBoundsException exception) {
      System.out.print("Invalid employee number: " + exception);
    }
  }

  /** Method updates an employee record in the file. */
  void writeEmployeeToRandomAccessFile(int employeeNumber) {
    try {
      // Position the file pointer.
      // Must add 4 bytes (int) at the start of the file used
      // for the number of records in the file.
      raf.seek(4 + (employeeNumber)* RECORD_LENGTH_BYTES);
      // Write employee data to file.
      writeEmployeeData(register.getEmployee(employeeNumber));
    }
    catch (IOException exception) {
      System.out.print("Error writing to file: " + exception);
    }
    catch (ArrayIndexOutOfBoundsException exception) {
      System.out.print("Invalid employee number: " + exception);
    }
  }

  /** Method writes all employees to the file. */
  void writeAllEmployeesToRandomAccessFile() {
    try {
      // Set the file pointer to the start of the file.
      raf.seek(0);

      // Write how many records will be stored in the file.
      int numOfEmployees = register.getNumOfEmployees();
      raf.writeInt(numOfEmployees);

      // Write data about each employee to the file.
      Employee[] employeeArray = register.getEmployeeArray();
      for (int i = 0; i < numOfEmployees; i++)
        writeEmployeeData(employeeArray[i]);
    }
    catch (IOException exception) {
      System.out.print("Error writing to the file: " + exception);
      System.exit(1);
    }
  }

  /** Method reads an employee record from the file. */
  void readEmployeeFromRandomAccessFile(int employeeNumber) {
    try {
      // Position the file pointer.
      // Must add 4 bytes (int) at the start of the file used
      // for number of records in the file.
      raf.seek(4 + (employeeNumber)* RECORD_LENGTH_BYTES);
      // Read the employee record at the current position.
      Employee employee = readEmployeeData();
      // Register in the personnel register.
      register.replaceEmployee(employeeNumber, employee);
    }
    catch (IOException exception) {
      System.out.print("Error reading from file: " + exception);
    }
  }
}