package game;    // File: game/GameBoard.java

public class GameBoard {
  // Constants:                                                // (1)
  public static final int ROW_COUNT = 6;
  public static final int COLUMN_COUNT = 7;
  public static final int CELL_COUNT = ROW_COUNT * COLUMN_COUNT;
  private static final int NO_VACANT_ROW = -1;

  // Immutable fields:                                         // (2)
  final IUserInterface ui;
  final IPieceIdentity[] pieceIdentities;
  final IPieceIdentity emptyCell;

  int moveCounter;                                             // (3)

  // private fields with state accessible through methods:     // (4)
  private IPieceIdentity[][] cellContents =
      new IPieceIdentity[COLUMN_COUNT][ROW_COUNT];
  private IPieceIdentity winner;

  // public constructors and methods:                          // (5)

  public GameBoard(IPieceIdentity[] identities, IPieceIdentity emptyCell) {
    this(identities, emptyCell, null);
  }

  public GameBoard(IPieceIdentity[] identities,
                   IPieceIdentity emptyCell, IUserInterface ui) {
    this.emptyCell = emptyCell;
    this.ui = ui;

    assert identities.length > 0;
    pieceIdentities = identities;
    reset();

    if (ui != null) {
      ui.setBoard(this);
    }
  }

  public void reset() {
    winner = null;
    moveCounter = 0;
    for (IPieceIdentity[] column : cellContents) {
      for (int rowIndex = 0; rowIndex < ROW_COUNT; ++rowIndex) {
        column[rowIndex] = emptyCell;
      }
    }
  }

  public IPieceIdentity pieceAt(int column, int row) {
    return cellContents[column][row];
  }

  public boolean hasWinner() {
    return winner != null;
  }

  public boolean isGameOver() {
    return (moveCounter >= CELL_COUNT) || hasWinner();
  }

  public void dropPieceDownColumn(IPieceIdentity piece, int column) {
    assert !isGameOver();
    int row = findBottommostVacantRow(column);
    cellContents[column][row] = piece;
    if (longestSequenceContainingCell(column, row) >= 4) {
      winner = piece;
    }
    ++moveCounter;
    reportToUserInterface();
  }

  public boolean isColumnFull(int column) {
    IPieceIdentity[] columnContents = cellContents[column];
    int indexOfTopmostCell = columnContents.length - 1;
    IPieceIdentity topmostCell = columnContents[indexOfTopmostCell];
    return topmostCell != emptyCell;
  }

  public boolean isValidColumnSelection(int column) {
    return validColumnIndex(column)
        && !isColumnFull(column);
  }

  public IPieceIdentity nextPiece() {
    return pieceIdentities[moveCounter % pieceIdentities.length];
  }

  protected void reportToUserInterface() {                     // (6)
    if (ui == null) {
      return;
    }

    ui.pieceWasDropped();
    if (!isGameOver()) {
      return;
    }

    if (hasWinner()) {
      ui.playerHasWon(winner);
    } else {
      ui.gameEndedInADraw();
    }
  }

  // private constructors and methods:                         // (7)

  private int findBottommostVacantRow(int column) {
    IPieceIdentity[] columnContents = cellContents[column];
    for (int row = 0; row < columnContents.length; ++row) {
      if (columnContents[row] == emptyCell) {
        return row;
      }
    }
    return NO_VACANT_ROW;
  }

  private int longestSequenceContainingCell(int column, int row) {
    int largestLength = 1;
    // Look for sequences of similar pieces in all directions.
    for (Direction direction : Direction.DIRECTIONS) {
      int piecesInALine = 1 + numOfConsecutivePieces(column, row, direction)
          + numOfConsecutivePieces(column, row, direction.opposite());
      largestLength = Math.max(largestLength, piecesInALine);
    }
    return largestLength;
  }

  private int numOfConsecutivePieces(int column, int row, Direction direction) {
    IPieceIdentity piece = pieceAt(column, row);
    int count = 0;
    while (true) {
      column += direction.columnStep;
      row += direction.rowStep;
      if (!positionContainsPiece(column, row, piece)) {
        break;
      }
      ++count;
    }
    return count;
  }

  private boolean validColumnIndex(int column) {
    return (column >= 0) && (column < COLUMN_COUNT);
  }

  private boolean validRowIndex(int row) {
    return (row >= 0) && (row < ROW_COUNT);
  }

  private boolean positionContainsPiece(int column, int row,
             IPieceIdentity piece) {
    return validColumnIndex(column)
        && validRowIndex(row)
        && (pieceAt(column, row) == piece);
  }
}