package museumPkg;

import java.awt.Container;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JTextField;

/**
 * GUI that can be customised to maintain MuseumItem objects.
 */
abstract public class AbstractItemFrame extends JFrame implements ActionListener {

  // Text fields.
  protected JLabel referenceLabel;
  protected JLabel nameLabel;
  protected JLabel descriptionLabel;
  protected JTextField referenceField;
  protected JTextField nameField;
  protected JTextField descriptionField;

  // Buttons
  protected JButton runButton;     // "Overridden" by subclasses.
  protected JButton exitButton;

  /**
   * Reference to main window to access the museum.
   */
  protected MainMuseumWindow mainWindow = null;

  /**
   * @param mainWindow Reference to main window where the Museum object
   * is to be found.
   */
  public AbstractItemFrame(MainMuseumWindow mainWindow) {
    this.mainWindow = mainWindow;

    //Create text fields.
    referenceLabel = new JLabel("Reference number:");
    nameLabel = new JLabel("Name:");
    descriptionLabel = new JLabel("Description:");
    referenceField = new JTextField();
    nameField = new JTextField();
    descriptionField = new JTextField();

    //Create buttons
    runButton = new JButton("");
    exitButton = new JButton("Exit");

    //Register listensers to the buttons
    referenceField.addActionListener(this);
    runButton.addActionListener(this);
    exitButton.addActionListener(this);

    //Create component hierarchy
    Container content = getContentPane();
    content.setLayout(new GridLayout(4, 2));
    content.add(referenceLabel);
    content.add(referenceField);
    content.add(nameLabel);
    content.add(nameField);
    content.add(descriptionLabel);
    content.add(descriptionField);
    content.add(runButton);
    content.add(exitButton);

    pack();
    setSize(200, 150);
    setVisible(true);
  }

  //Implements the ActionListener interface.
  /**
   * Called when a button is clicked.
   * Catches MuseumItemException and shows it in a new window.
   */
  public void actionPerformed(ActionEvent e) {
    try {
      if (e.getSource() == runButton)
        doAction();
      else if (e.getSource() == exitButton)
        removeFrame();
      else if (e.getSource() == referenceField)
        findItemInMuseum();
    } catch (MuseumItemException mie) {
      new ErrorReportFrame(mie.getDescription());
    }
  }

  /**
   * Method can be overridden to implement the action for the "runButton".
   * Does nothing in this class.
   * @exception MuseumItemException If an error occurs in any of the methods
   * that are called.
   */
  void doAction() throws MuseumItemException { }

  /**
   * Method implements the action for the "exitButton".
   */
  void removeFrame() {
    mainWindow.setNumOfItems();
    setVisible(false);
    dispose();
  }

  /**
   * Method can be overridden to implement the action when Enter key
   * is typed in a text field.
   * Does nothing in this class.
   * @exception MuseumItemException If an error occurs its code.
   */
  void findItemInMuseum() throws MuseumItemException { }

  /**
   * Returns the integer corresponding to the string in the reference field.
   * @exception MuseumItemException If the string is not a valid integer.
   */
  int getItemReferenceNum() throws MuseumItemException {
    try {
      return Integer.parseInt(referenceField.getText());
    } catch (NumberFormatException e) {
      throw new MuseumItemException("Invalid integer!");
    }
  }

  /**
   * Returns the string in the name field.
   * @exception MuseumItemException If the name field is empty.
   */
  String getItemName() throws MuseumItemException {
    String name = nameField.getText();
    if (name.length() == 0)
      throw new MuseumItemException("Name missing!");
    else
      return name;
  }

  /**
   * Returns the string in the description field.
   * @exception MuseumItemException If the description field is empty.
   */
  String getItemDescription() throws MuseumItemException {
    String description = descriptionField.getText();
    if (description.length() == 0)
      throw new MuseumItemException("Description missing!");
    else
      return description;
  }
}
